```@meta
CurrentModule = MosaikAPIv3
```

# MosaikAPIv3

[MosaikAPIv3](https://gitlab.com/mosaik/api/MosaikAPIv3.jl) allows you to connect tools written in Julia
to the co-simulation tool [mosaik](https://mosaik.offis.de/) written in Python.
This will allow users of mosaik to integrate your tool into their simulation.

!!! info "Why is this version 3?"

    While this is the first adapter connecting Julia to mosaik,
    this adapter implement the current version of the mosaik API, which is version 3.
    It is compatible with all minor releases of mosaik 3,
    and will potentially remain compatible with higher versions of mosaik,
    even if those also provide higher versions of the API.

## A few words on mosaik

mosaik allows you to couple simulation tools together.
It will orchestrate when these tools get to perform their calculations and organize the data flow between them.
The user of mosaik creates a *scenario* which describes which tools to use and how to connect them.

In the context of mosaik, each tool is called a *simulator*.
Simulators are run at steps that they decide for themselves and can also be configured to run whenever they receive new input.
To structure the inputs and outputs, each simulator provides *models* and these models in turn specify *attributes* (usually *attrs* for short).
During scenario creation, both the user and your simulator get to create instances of these models, called *entities*.
The user then connects these entities to other entities in their scenario, specifying which attributes of the first entity should send data to which attributes of the second entity.
Entities often represent concrete objects in your simulation (for example, each entity couly represent a single PV power plant, with your simulator running calculations for many power plants at once),
but entities can also be used for more abstract purposes (for example, to provide different endpoints to an electric power market depending on whether the energy is produced sustainably or not).

It is important to note that mosaik will always call your entire simulator, not individual entities.
This means that you can and must organize communication between the entities in your simulator yourself.

For further information, details about the mosaik environment and tutorials to help you get started with mosaik, please visit the official [documentation](https://mosaik.readthedocs.io/en/latest/index.html). 
Examples for simple Julia simulators can be found [here](https://gitlab.com/mosaik/api/mosaik-api-julia/-/blob/main/test/input_sim_julia.jl?ref_type=heads) and [here](https://gitlab.com/mosaik/api/mosaik-api-julia/-/blob/main/test/output_sim_julia.jl?ref_type=heads).


## Overview about what needs to be implemented by you

During scenario setup, your simulator gets an initialization call (see [`init`](@ref))
and must return some *meta* information about itself to mosaik (see [`meta`](@ref)).
Then, the user can create entities in your simulator (see [`create!`](@ref)).

When the scenario is run, your simulator will occassionally be asked to perform a *step*.
It will be told the current (simulation-internal) time and provided with the input data for all its entities (see [`step!`](@ref)).
Directly afterwards, it will then be asked for its outputs (see [`get_data`](@ref)).

If needed, your simulator can also perform some final setup immediately before the simulation starts (see [`setup_done!`](@ref))
or some tear-down at the end (see [`finalize!`](@ref)).

When implementing this API, you will usually first [create a simulator type](@ref create-type),
then implement the mosaik methods for it and finally start your simulator by calling [`run_server`](@ref) or [`run_client`](@ref).


## [Creating a simulator type](@id create-type)

The first step to integrating your tool with mosaik is to create a type to hold its state.
This type must be a subtype of the abstract type `Simulator`:

```@docs
Simulator
```

In this type, you will usually store a list of user-created entities together with their associated data.
It will also store all other information that your simulator needs to preserve over the course of the simulation.
In this document, we will refer to your simulator type as `Sim` whereas `Simulator` is the abstract base class provided by MosaikAPIv3.

A common simulator type might look like this:

```julia
mutable struct Sim <: MosaikAPIv3.Simulator
    sim_id::String
    entities::Dict{String,Entity}
    mosaik_remote::MosaikRemote
    ...
end
```

Here, `Entity` is whatever type you use to represent your simulator's entities.
(It doesn't have to be a custom type if your entities are simple enough.)


## [Implementing the mosaik methods](@id mosaik-methods)

Having created your type, you need to implement the functions of the mosaik API for it.
Each of them (except for [`init`](@ref), see below) will be called with your `Sim` as the first parameter,
allowing Julia to dispatch to your methods of these functions.

A word on convention: In Python, the mosaik methods are called `init`, `create`, `step`, `get_data`
(and `setup_done` and `finalize`, if needed).
In Julia, we use the same names but add an exclamation point at the end if the method is expected to change the simulator,
following Julia conventions.
We also add the method [`meta`](@ref), for reasons explained below.

### `init` and `meta`

The first method to write is `init`.
Unlike the other methods (and unlike its Python counterpart), this method will not receive the simulator instance as its first parameter,
because that instance has not been created yet.
Instead, the first parameter will be your simulator type, which we call `Sim` in this tutorial.
This type's type is `Type{Sim}`, so this is what you need to specify as the type of the first parameter for Julia's dispatching to work.
Unlike in Python, `init` should create and return an instance of your simulator type.
Therefore, it doesn't return the simulator's meta, which is handled by the method [`meta`](@ref), instead.

```@docs
init
```

Directly after initialization, MosaikAPIv3 will call `meta` on your simulator.
The result will be returned to mosaik as the response to its `init` call.

```@docs
meta
MosaikMeta
SimulatorType
Model
```

### `create!`

Whenever the user creates entities in their scenario, `create!` will be called. Below, a common way to implement this function into a simulator is displayed.

```julia
function MosaikAPIv3.create!(sim::Sim, num::Integer, model::String; kwargs...)
    previous_count = length(sim.entities)
    new_entities = []
    for i in (previous_count + 1):(previous_count + num)
        eid = "$model-$i"
        sim.entities[eid] = Entity(...)
        push!(new_entities, CreateResult(;
            eid = eid,
            type = model,
        ))
    end
    return new_entities
end
```

```@docs
create!
CreateResult
```

### `step!` and `get_data`

During the simulation, mosaik will call `step!` whenever your simulator should perform a step.
```@docs
step!
```

After each step, `get_data` is called, unless your simulator's outputs are not connected to anything in the user's scenario script (in which case `get_data` is never called).

```@docs
get_data
```

Analogous to the Python implementation, `get_data` receives its data in the form of a `OutputRequest` and sends it back to mosaik as an `OutputData`.

```@docs
OutputRequest
OutputData
```

### Optional methods: `setup_done!` and `finalize!`

Your simulator may optionally implement `setup_done!` and `finalize!`,
which will be called by mosaik directly before and after (respectively) the simulation runs.
In particular, when `setup_done!` is called, all entities in the simulation have already been created and connected.
The default implementations of these methods do nothing.

```@docs
setup_done!
finalize!
```


## [Starting your simulator](@id start-simulator)

When you have implemented these functions for your simulator, you can finally start it.
This is done using either the `run_server` function or the `run_client` function.
In the former case, your simulator awaits an incoming connection from the user's scenario script,
in the latter case, your simulator must be started by the user's scenario script and then establishes a connection itself.
These two functions correspond to the `"connect"` and `"cmd"` options in the user's `SimConfig`
([see this page](https://mosaik.readthedocs.io/en/latest/simmanager.html) of the mosaik documentation.)

Note that unlike the mosaik methods in the last section, you don't implement these methods.
Instead, you just call them in your Julia script's entrypoint.

```@docs
run_server
run_client
```

## [Making asynchronous requests](@id async-requests)

Occasionally, you will want to make additional requests to mosaik that are not covered by the usual `step!`-`get_data` loop.
For this, mosaik provides a couple of so-called *asynchronous requests*.
To call these methods from your Julia adapter, you need to store the `MosaikRemote` object that is passed to your [`init`](@ref) method.
However, as long as you have access to this object, you may call these functions even outside of the mosaik methods, for example from a separate thread.

```@docs
MosaikRemote
```

The following asynchronous requests are supported by the Julia adapter.
We don't support the `set_data` and `get_data` asynchronous requests,
as they are already deprecated in mosaik.

```@docs
get_progress
get_related_entities
set_event
```

