module MosaikAPIv3

export OutputRequest,
    OutputData,
    run_server,
    sim,
    CreateResult,
    init,
    create!,
    setup_done!,
    step!,
    get_data,
    meta,
    finalize!,
    Simulator,
    MosaikRemote,
    run_client,
    Model,
    MosaikMeta

using Sockets
using DocStringExtensions
import JSON3
import StructTypes
import Base.show

# custom data types

"""The requested outputs for get_data. For each entity where data is
needed, the required attributes are listed."""
const OutputRequest = Dict{String,Vector{String}}

"""The format of output data as returned by ``get_data``"""
const OutputData = Dict{String,Dict{String,Any}}

"""The format of input data used as a parameter by the ``step`` function."""
const InputData = Dict{String, Dict{String, Dict{String, Any}}}

"""The super type ::Simulator that every sim type is inherited from."""
abstract type Simulator end


# structs

"""
Information about an individual entity returned by [`create!`](@ref).
$(TYPEDFIELDS)
"""
@kwdef struct CreateResult
    eid::String
    type::String
    rel::Vector{String} = []
    children::Vector{CreateResult} = []
    extra_info::Any = nothing
end

"""
The type of a simulator, specifying how it is stepped.
$(TYPEDFIELDS)
"""
@enum SimulatorType begin
    time_based
    event_based
    hybrid
end

StructTypes.StructType(::Type{SimulatorType}) = StructTypes.StringType()
function Base.string(t::SimulatorType)
    if t == time_based
        "time-based"
    elseif t == event_based
        "event-based"
    elseif t == hybrid
        "hybrid"
    else
        error("not a SimulatorType")
    end
end

"""
A struct describing an individual model in the [`MosaikMeta`](@ref).

# Fields
-   `public::Bool`: Whether entities of this model can be created by the user in their
    scenario.
-   `params::Vector{Symbol}`: The names of the params that may be specified when creating
    entities of this model. (Their types and whether they're optional needs to be documented
    separately.)
-   `attrs::Union{Vector{Symbol},Nothing}`: The attributes of this model. You can also
    specify `nothing` if you give sufficiently many of `trigger`, `non_trigger`,
    `persistent`, and `non_persistent`, instead.
-   `any_inputs::Bool`: Whether this model simulator accepts input on all attributes. (This
    is useful for databases and such.)
-   `trigger::Union{Vector{Symbol},Nothing}`: The trigger input attributes, if any.
-   `non_trigger::Union{Vector{Symbol},Nothing}`: The non-trigger input attributes, if any.
-   `persistent::Union{Vector{Symbol},Nothing}`: The persistent output attributes, if any.
-   `non_persistent::Union{Vector{Symbol},Nothing}`: The non-persistent output attributes,
    if any.
"""
@kwdef struct Model
    public::Bool = true
    params::Vector{Symbol}
    attrs::Union{Vector{Symbol},Nothing}
    any_inputs::Bool = false
    trigger::Union{Vector{Symbol},Nothing} = nothing
    non_trigger::Union{Vector{Symbol},Nothing} = nothing
    persistent::Union{Vector{Symbol},Nothing} = nothing
    non_persistent::Union{Vector{Symbol},Nothing} = nothing
end

StructTypes.StructType(::Type{Model}) = StructTypes.CustomStruct()

function StructTypes.lower(model::Model) 
    model_dict = Dict(
        "public" => model.public,
        "params" => model.params,
        "any_inputs" => model.any_inputs,
    )
    optionals = Dict(
        :attrs => "attrs",
        :trigger => "trigger",
        :non_trigger => "non-trigger",
        :persistent => "persistent",
        :non_persistent => "non-persistent",
    )
    for (struct_field, dict_field) in optionals
        value = getfield(model, struct_field)
        if value !== nothing
            model_dict[dict_field] = value
        end
    end
    model_dict
end

"""
A struct containig the following fields:

-   `type::SimulatorType`: The type of your simulator, see [`SimulatorType`](@ref).
-   `models::Dict{Symbor,Model}`: A dictionary of models, mapping each models name to a 
    [`Model`](@ref) structure describing it.
-   `extra_methods::Vector{Function}`: A list of the extra methods provided by this
    simulator. The method's name will be extracted and passed on to mosaik.
"""
@kwdef struct MosaikMeta
    type::SimulatorType
    models::Dict{Symbol,Model}
    extra_methods::Vector{Function} = []
end
StructTypes.StructType(::Type{MosaikMeta}) = StructTypes.CustomStruct()
StructTypes.lower(meta::MosaikMeta) = Dict(
    :api_version => "3.0",
    :type => meta.type,
    :models => meta.models,
    :extra_methods => map(Symbol, meta.extra_methods),
)

"""
An opaque type representing the connection of your simulator to mosaik. You will need to
supply this when you want to make asynchronous requests to mosaik.
"""
mutable struct MosaikRemote
    """The socket connection to mosaik."""
    sock::TCPSocket

    """A dictionary responsible for creating and storing separate channels for outgoing,
    asynchronous requests."""
    outgoing_requests::Dict{Int,Channel}

    """The id counter used to distribute unique identifiers."""
    id_counter::Int64

    """The channel used to store and retreive requests coming from mosaik."""
    main_channel::Channel

    function MosaikRemote(sock::TCPSocket)
        new(sock, Dict{Int,Channel}(), -1, Channel(1))
    end
end


# private functions

"""
    _init_sim(sim_type::Type{<:Simulator}, sock::TCPSocket, extra_methods::Dict)

Initialize a simulator of type `sim_type` using a TCP socket contained in `mosaik_remote` and update
 `extra_methods` with additional methods provided by the simulator.

# Arguments
- `sim_type::Type{<:Simulator}`: The type of simulator to initialize.
- `mosaik_remote::MosaikRemote`: The connection struct containing the TCP socket.
- `extra_methods::Dict`: A dictionary containing extra methods provided by the simulator.

# Returns
- `sim::Simulator`: The initialized simulator object.
"""
function _init_sim(
    sim_type::Type{<:Simulator},
    mosaik_remote::MosaikRemote,
    extra_methods::Dict,
)
    sim = nothing
    msg_id, msg = take!(mosaik_remote.main_channel)
    func, args, kwargs = msg
    if func == :init
        sim_id, = args
        time_resolution = Float64(pop!(kwargs, :time_resolution))
        sim = init(sim_type, mosaik_remote, sim_id, time_resolution; kwargs...)
        result = meta(sim)
        if hasproperty(result, :extra_methods)
            for method in result.extra_methods
                extra_methods[Symbol(method)] = method
            end
        end
        _write_mosaik_message(mosaik_remote.sock, [1, msg_id, result])
    else
        if typeof(func) == Symbol
            throw("Mosaik did not send the the init-message it was expected to.
               Instead, it sent $func.")
        else
            throw("Received non-readable message from mosaik: $func.")
        end
    end
    sim
end


"""
    _get_id()

Generate a unique identifier.

# Arguments
- `mosaik_remote::MosaikRemote`: Structure to manage the connection to mosaik and store shared variables.

# Returns
A unique identifier.
"""
function _get_id(mosaik_remote::MosaikRemote)
    mosaik_remote.id_counter = mosaik_remote.id_counter + 1
    mosaik_remote.id_counter
end


"""
    _process_async_request(msg_payload::Vector, sock::TCPSocket)

Process an asynchronous request by sending it over a TCP socket.

# Arguments
- `msg_payload::Vector`: Payload of the message.
- `mosaik_remote::MosaikRemote`: Structure to manage the connection to mosaik and store shared variables.

# Returns
A Channel for handling the response of the asynchronous request.
"""
function _process_async_request(msg_payload::Vector, mosaik_remote::MosaikRemote)
    id = _get_id(mosaik_remote)
    msg = [0, id, msg_payload]
    mosaik_remote.outgoing_requests[id] = Channel(1)
    _write_mosaik_message(mosaik_remote.sock, msg)
    mosaik_remote.outgoing_requests[id]
end


"""
    _read_mosaik_message(sock)

Read and parse a Mosaik message from a TCP socket.

# Arguments
- `sock`: TCP socket for communication.

# Returns
A tuple representing the parsed Mosaik message.

# Throws
- `EOFError` if the socket is has no more data
"""
function _read_mosaik_message(sock)
    length = ntoh(read(sock, Int32))
    msg_json = read(sock, length)
    JSON3.read(msg_json, Tuple{Int,Int,Any})
end


"""
    _convert_to_mosaik_request(value::Any)

Convert a generic value to a specific Mosaik request type.

# Arguments
- `value::Any`: Generic value to be converted.

# Returns
A tuple representing the specific Mosaik request type.

# Error
- Throws an error if conversion fails.
"""
function _convert_to_mosaik_request(value::Any)::Tuple{Symbol,Vector{Any},Dict{Symbol,Any}}
    try
        symbol_value, vector_value, dict_value = value
        symbol_value = Symbol(symbol_value)
        vector_value = Vector{Any}(vector_value)
        dict_value = Dict{Symbol,Any}(Symbol(k) => v for (k, v) in dict_value)

        specific_type_value = (symbol_value, vector_value, dict_value)

        return specific_type_value
    catch e
        @error "In function 'convert to mosaik request'" exception = (e, catch_backtrace())
        error("Failed to convert $value into a mosaik request.")
    end
end


"""
    _write_mosaik_message(sock, msg)

Write a Mosaik message to a TCP socket.

# Arguments
- `sock`: TCP socket for communication.
- `msg`: Mosaik message to be sent.

# Returns
Nothing.
"""
function _write_mosaik_message(sock, msg)
    msg_json = JSON3.write(msg)
    write(sock, hton(Int32(length(msg_json))))
    write(sock, msg_json)
end

# asynchronous requests

"""
    get_progress(mosaik_remote::MosaikRemote)

Requests the simulation progress from mosaik by sending the async request 'get_progress'.

# Arguments
- `mosaik_remote::MosaikRemote`: Mosaik remote object.

# Returns
The progress of the simulation in percent.
"""
function get_progress(mosaik_remote::MosaikRemote)
    msg_payload = ["get_progress", [], Dict()]
    take!(_process_async_request(msg_payload, mosaik_remote))
end


"""
    get_related_entities(mosaik_remote::MosaikRemote, entities::Vector{String})

Request related entities for a given vector of entity IDs from the Mosaik remote.

# Arguments
- `mosaik_remote::MosaikRemote`: Mosaik remote object.
- `entities::Vector{String}`: Vector of entity IDs.

# Returns
The related entities.
"""
function get_related_entities(mosaik_remote::MosaikRemote, entities::Vector{String})
    wrapped_entities = [entities]
    msg_payload = ["get_related_entities", wrapped_entities, Dict()]
    take!(_process_async_request(msg_payload, mosaik_remote))
end


"""
    set_event(mosaik_remote::MosaikRemote, step::Int)

Notify the Mosaik remote about a simulation event at a specific step.

# Arguments
- `mosaik_remote::MosaikRemote`: Mosaik remote object.
- `step::Int`: Simulation step.

# Returns
`nothing` once mosaik has acknowledged the `set_event` call.
"""
function set_event(mosaik_remote::MosaikRemote, step::Int)
    wrapped_step = [step]
    msg_payload = ["set_event", wrapped_step, Dict()]
    take!(_process_async_request(msg_payload, mosaik_remote))
end


# core functions to be used by the created simulators


"""
    init(::Type{Sim}, mosaik_remote::MosaikRemote, sim_id::String, time_resolution::Float64; kwargs...)::Sim

This function will be called once the `init` call from mosaik arrives. `sim_type` will be
the subtype of [`Simulator`](@ref) that you specified in your call to [`run_client`](@ref)
or [`run_server`](@ref), so that Julia can dispatch based on that type. (In other words,
you should explicitly give its type as `Type{Sim}`, where `Sim` is your concrete simulator
type.)

This function should construct and initialize an instance of your type, based on the given
arguments, and return it. Note that this is different from the Python version, where `init`
is called on an already-constructed simulator object to (further) initialize it, and returns
the simulator's meta. See [`meta`](@ref) for information on how to return your meta here.

# Arguments
-   `sim_type::Type{Sim}`: The *type* of your simulator (a subtype of `Simulator`), used for
    dispatch.
-   `mosaik_remote::MosaikRemote`: The [`MosaikRemote`](@ref) for this simulator. You need
    to save it in your simulator if you want to make requests back to mosaik, otherwise, you
    can ignore it.
-   `sim_id::String`: The ID assigned to this simulator in the user's scenario.
-   `time_resolution::Float64`: The time resolution for the scenario. Each step of mosaik
    time corresponds to `time_resolution` seconds.
-   `kwargs`: Further keyword arguments that the user specified when starting your
    simulator in their scenario script. You should document which keyword arguments your
    simulator supports in its documentation.

# Returns
Your simulator. It should be an instance of your simulator type. 
"""
function init end

"""
    meta(sim::Sim)

This function will be called right after [`init`](@ref) and must create a
[`MosaikMeta`](@ref) that contains the meta of the simulator.

# Arguments
- `sim::Sim`: Your simulator instance.

# Returns
A `MosaikMeta` describing your simulator.
"""
function meta end

"""
    setup_done!(sim::Sim)

A function to indicate that setup is done for the simulator.

# Arguments
- `sim::Sim`: Your simulator instance.

# Returns
Nothing.
"""
function setup_done!(::Simulator) end

"""
    step!(sim::Sim, time::Int, inputs::Dict{String, Any}, max_advance::Int)

A function to execute a single simulator step. 
It is mandatory to implement this function into a simulator.

# Arguments
- `sim::Sim`: Your simulator instance.
- `time::Int`: The current simulation time.
- `inputs::Dict{String, Any}`: A dictionary containing input data for the simulator.
- `max_advance::Int`: The maximum time advance allowed for this step.

# Returns
An integer specifying the next step at which the simulator wants to be called, or `nothing`
if no step should be scheduled. (The simulator will still be called again if new input on
a trigger attribute arrives.)
"""
function step! end

"""
    create!(sim::Sim, num::Int64, model::String)

A function to create simulator entities.

# Arguments
- `sim::Sim`: Your simulator instance.
- `num::Int64`: The number of entities to create.
- `model::String`: The model type of the entities to create.

# Returns
A list of the eids of all newly created entities as type Vector{Any}.
"""
function create! end


"""
    get_data

Retrieves data from the simulator to send it to mosaik

# Returns
The requested data as a type :OutputData.
"""
function get_data end

"""
    finalize!(sim::Sim)

Perform finalizing work at the end of the simulation. (For example, you might close a
database connection here.)

# Arguments
- `sim::Sim`: Your simulator instance.

# Returns
Nothing.
"""
function finalize!(::Simulator) end


"""
    run_server(sim_type::Type{<:Simulator}; port=2000)

Start a Mosaik server for the specified simulator type.

# Arguments
- `sim_type::Type{<:Simulator}`: Type of the simulator to run.
- `port::Int`: Port number to listen on (default is 2000).

# Returns
Nothing.
"""
function run_server(sim_type::Type{<:Simulator}; port = 2000)
    server = listen(port)
    sock = accept(server)
    close(server)
    run(sim_type, sock)
end


"""
    run_client(sim_type::Type{<:Simulator}; host, port)

Connect to a Mosaik server as a client using the specified simulator.

# Arguments
- `sim::Simulator`: Simulator instance.
- `host::String`: Hostname or IP address of the server.
- `port::Int`: Port number to connect to.

# Returns
Nothing.
"""
function run_client(sim_type::Type{<:Simulator}; host, port)
    sock = connect(host, port)
    run(sim_type, sock)
end


# core mosaik api functions

"""
    run(sim_type::Type{<:Simulator}, sock::TCPSocket)

Run the main Mosaik loop for communication with a simulator.

# Arguments
- `sim_type::Type{<:Simulator}`: Type of the simulator.
- `sock::TCPSocket`: TCP socket for communication.

# Returns
Nothing.
"""
function run(sim_type::Type{<:Simulator}, sock::TCPSocket)
    try
        mosaik_remote = MosaikRemote(sock)
        message_loop_task = @task _mosaik_message_loop(mosaik_remote)
        sim_loop_task = @task try
            _main_mosaik_loop(sim_type, mosaik_remote)
        catch
            @async Base.throwto(message_loop_task, :simulator_error)
            rethrow()
        end
        schedule(message_loop_task)
        schedule(sim_loop_task)
        try
            wait(message_loop_task)
        catch e
            if !(e isa TaskFailedException && e.task.result == :simulator_error)
                @error "Exception in message loop" exception = (e, catch_backtrace())
            end
        end
        try
            wait(sim_loop_task)
        catch e
            @error "Exception in simulator" exception = (e, catch_backtrace())
        end
    finally
        @info "Closing socket"
        close(sock)
    end
end


function _mosaik_message_loop(mosaik_remote::MosaikRemote)
    while true
        try
            type, msg_id, msg = _read_mosaik_message(mosaik_remote.sock)
            if type == 0
                decoded_msg = _convert_to_mosaik_request(msg)
                put!(mosaik_remote.main_channel, (msg_id, decoded_msg))
            elseif type == 1
                put!(mosaik_remote.outgoing_requests[msg_id], msg)
            else
                throw("Unknown message type from mosaik received: $type")
            end
        catch e
            if e isa EOFError
                close(mosaik_remote.main_channel)
                break
            else
                rethrow()
            end
        end
    end
end


"""
    _main_mosaik_loop(sim_type::Type{<:Simulator}, sock::TCPSocket)

Main loop for handling Mosaik communication with the specified simulator type.

# Arguments
- `sim_type::Type{<:Simulator}`: Type of the simulator.
- `sock::TCPSocket`: TCP socket for communication.

# Returns
Nothing.
"""
function _main_mosaik_loop(sim_type::Type{<:Simulator}, mosaik_remote::MosaikRemote)
    extra_methods = Dict()
    sim = _init_sim(sim_type, mosaik_remote, extra_methods)
    for (msg_id, msg) in mosaik_remote.main_channel
        func, args, kwargs = msg
        if func == :create
            result = create!(sim, args...; kwargs...)
        elseif func == :setup_done
            result = setup_done!(sim, args...; kwargs...)
        elseif func == :step
            result = step!(sim, args...; kwargs...)
        elseif func == :get_data
            output_request = OutputRequest()
            for dict in args
                for (key, values) in dict
                    converted_values = String.(values)
                    output_request[key] = converted_values
                end
            end
            result = get_data(sim, output_request; kwargs...)
        elseif func == :stop
            finalize!(sim)
            break
        else
            throw("Unknown mosaik function $func")
        end
        _write_mosaik_message(mosaik_remote.sock, [1, msg_id, result])
    end
end


end