using Test
using Logging

"""
    gather(funcs...)

Schedule each element of `funcs` as a task. If one of them fails, cancel the rest.
"""
function gather(funcs...)
    @debug "Gathering $(length(funcs)) tasks now"
    tasks = map(
        (x -> @task begin
            i, func = x
            try
                func()
            catch e
                @debug "Got exception in task $i" exception = (e, catch_backtrace())
                # As we are use `include` in our tasks below, the `:stop` that we throw
                # in might be wrapped in a `LoadError` once it reaches us here.
                if e == :stop || (e isa LoadError && e.error == :stop)
                    @debug "Task $i has been cancelled"
                else
                    cancel_all_but(i)
                    rethrow()
                end
            end
        end),
        enumerate(funcs),
    )
    function cancel_all_but(except_i)
        @debug "Cancel all but task $except_i"
        for (i, task) in enumerate(tasks)
            if i == except_i
                continue
            end
            @debug "Cancelling task $i"
            if !(istaskdone(task) || istaskfailed(task))
                @async Base.throwto(task, :stop)
            end
        end
    end
    @debug "Starting tasks"
    for task in tasks
        schedule(task)
    end
    for task in tasks
        wait(task)
    end
end

@testset "Integration tests" begin
    # Can be turned on by changing `begin` on one of the tests to
    # `with_logger(debug_logger) do`.
    debug_logger = ConsoleLogger(stderr, Logging.Debug)

    # These should probably be implemented using some other framework as Test is intended
    # for unit tests.
    @test begin
        println("Current working directory: ", pwd())  # Print current working directory
        gather(
            () -> include("input_sim_julia.jl"),
            () -> begin
                sleep(0.1)
                run(`python3 --version`)
                run(`python3 python/test_input_julia.py`)
            end,
        )
        # Just test whether simulation completes. If we get here, return success
        true
    end

    @test begin
        gather(
            () -> include("output_sim_julia.jl"),
            () -> begin
                sleep(0.1)  # To make sure that Python's not too quick.
                run(`python3 python/test_output_julia.py`)
            end,
        )
        # Just test whether simulation completes. If we get here, return success
        true
    end

    @test begin
        gather(
            () -> include("get_progress_julia.jl"),
            () -> begin
                sleep(0.1)  # To make sure that Python's not too quick.
                run(`python3 python/test_get_progress_julia.py`)
            end,
        )
        # Just test whether simulation completes. If we get here, return success
        true
    end

    @test begin
        gather(
            () -> include("get_related_entities_julia.jl"),
            () -> begin
                sleep(0.1)  # To make sure that Python's not too quick.
                run(`python3 python/test_get_related_entities_julia.py`)
            end,
        )
        # Just test whether simulation completes. If we get here, return success
        true
    end

    @test begin
        gather(
            () -> include("set_event_julia.jl"),
            () -> begin
                sleep(0.1)  # To make sure that Python's not too quick.
                run(`python3 python/test_set_event_julia.py`)
            end,
        )
        # Just test whether simulation completes. If we get here, return success
        true
    end
end